# Bu araç @keyiflerolsun tarafından | @KekikAkademi için yazılmıştır.

from rich    import print
from decimal import Decimal

def sikici_yaz(sozluk: dict, girinti: int = None, kademe: int = Decimal("Infinity")):
    """
    Sözlük Tipindeki Veriyi Nesne Formatında Ekrana Yazdırır.
        `https://github.com/LonamiWebs/Telethon/blob/master/telethon/_misc/helpers.py#L192`

    Parametreler
    ------------
        >>> sozluk  (dict) # Yazılması İstenen Veri
        >>> girinti (int)  # Kaç Tab Boşluk Bıraksın
        >>> kademe  (int)  # En Fazla Kaç Kademe Göstersin

    Kullanım
    --------
    ```
    sikici_yaz(sozluk={"Merhaba": "Dünya"}, girinti=0, kademe=1)
    ```

    Dönüt
    -----
        >>> None
    """
    kademe -= 1
    if kademe < 0:
        return "..."

    to_d = getattr(sozluk, "to_dict", None)
    if callable(to_d):
        sozluk = to_d()

    if girinti is None:
        if isinstance(sozluk, dict):
            return f"{sozluk.get('_', 'Nesne')}({', '.join(f'{k}={sikici_yaz(v, girinti, kademe)}' for k, v in sozluk.items() if k != '_')})"

        elif isinstance(sozluk, (str, bytes)) or not hasattr(sozluk, "__iter__"):
            return repr(sozluk)

        else:
            return f"[{', '.join(sikici_yaz(x, girinti, kademe) for x in sozluk)}]"
    else:
        sonuc = []

        if isinstance(sozluk, dict):
            sonuc.extend((sozluk.get("_", "Nesne"), "("))
            if sozluk:
                sonuc.append("\n")
                girinti += 1
                for k, v in sozluk.items():
                    if k == "_":
                        continue
                    sonuc.extend(
                        ("\t" * girinti, k, "=", sikici_yaz(v, girinti, kademe), ",\n")
                    )

                sonuc.pop()  # son ',\n'
                girinti -= 1
                sonuc.extend(("\n", "\t" * girinti))
            sonuc.append(")")

        elif isinstance(sozluk, (str, bytes)) or not hasattr(sozluk, "__iter__"):
            sonuc.append(repr(sozluk))

        else:
            sonuc.append("[\n")
            girinti += 1
            for x in sozluk:
                sonuc.extend(("\t" * girinti, sikici_yaz(x, girinti, kademe), ",\n"))
            girinti -= 1
            sonuc.extend(("\t" * girinti, "]"))

        return "".join(sonuc)

class VeriTipiHatasi(Exception):
    """
    Veri Tipi Hatası

    Parametreler
    -----------
        >>> veri        # Hataya Sebep Veri
        >>> mesaj (str) # Hata Sonucu Mesajı
    """
    def __init__(self, veri, mesaj:str="Sözlük Tipinde Bir Veri Değil!"):
        self.veri  = veri
        self.mesaj = mesaj
        super().__init__(self.mesaj)

    def __str__(self):
        return f"\n\n{self.veri} -> {self.mesaj}"

class Nesne(object):
    """
    Sözlük Veri Tipini Python Nesnesine Çevirir.

    Parametreler
    -----------
        >>> sozluk (dict) # Nesneye Çevrilmek İstenen Sözlük Tipli Veri

    Nitelikler
    ----------
        >>> .sozluk -> dict:
        Verilen Sözlüğün Kendisini Döndürür.

        >>> # Sözlüğün Anahtarları -> Sözlüğün Değeri:
        Anahtarın Değerini Döndürür.

    Metodlar
    ----------
        >>> .gorsel(girinti:int, kademe:int) -> None:
        >>> # Oluşan Nesneyi Görsel Biçimde Konsola Yazdırır.

        >>> # Örnek; <---------------------------------- #
        >>> .gorsel(girinti=0, kademe=1)

        >>> # ------------------------------------------ #
        >>> .anahtarlar(nesne:Nesne) -> list:
        >>> ['isSuccess', 'statusCode', 'error', 'result', 'headers']

        >>> # ------------------------------------------ #
        >>> .sozluk -> dict:
        >>> # girdinin son hali
    """
    def __repr__(self) -> str:
        mesaj = f"{__class__.__name__}("
        for anahtar in self.__sozluk.keys():
            mesaj += f"{anahtar}={self.__dict__[anahtar]}, "

        return f"{mesaj.rstrip(', ')})"

    def __str__(self) -> str:
        return repr(self)

    def __init__(self, sozluk:dict):
        self.__sozluk = sozluk

        if not isinstance(self.__sozluk, dict):
            raise VeriTipiHatasi(self.__sozluk)

        # https://stackoverflow.com/questions/1305532/how-to-convert-a-nested-python-dict-to-object
        for anahtar, deger in self.__sozluk.items():
            if isinstance(deger, (tuple, list, set, frozenset)):
                object.__setattr__(
                    self,
                    anahtar,
                    [
                        Nesne(veri) if isinstance(veri, dict) else veri
                          for veri in deger
                    ]
                )
            else:
                object.__setattr__(
                    self,
                    anahtar,
                    Nesne(deger) if isinstance(deger, dict) else deger
                )

    # * Sözlük Benzeri -> erişim / güncellemeler
    def __getitem__(self, anahtar):
        deger = self.__dict__[anahtar]
        if isinstance(deger, dict):  # * Yinelemeli olarak alt sözlükleri nesne olarak görüntüle
            deger = Nesne(deger)
        return deger

    def __setitem__(self, anahtar, deger):
        self.__dict__[anahtar] = deger

    def __delitem__(self, anahtar):
        del self.__dict__[anahtar]

    # * Nesne Benzeri -> erişim / güncellemeler
    def __getattr__(self, anahtar):
        return self[anahtar]

    def __setattr__(self, anahtar, deger):
        self[anahtar] = deger

    def __delattr__(self, anahtar):
        del self[anahtar]

    @property
    def sozluk(self) -> dict:
        """Girdinin Son Hali"""
        veri = self.__dict__.copy()
        del veri["_Nesne__sozluk"]

        return veri

    @staticmethod
    def anahtarlar(veri) -> list | None:
        """
        Parametre olarak verilen Nesne Verisinin Anahtarlarını döndürür.

        Parametreler
        ------------
            >>> veri (Nesne) # Anahtarları İstenen Nesne Verisi

        Dönüt
        -----
        >>> list | None
        """
        if isinstance(veri, Nesne):
            veri = veri.__dict__.copy()
            del veri["_Nesne__sozluk"]

            if sozluk := veri:
                if isinstance(sozluk, dict):
                    return list(veri.keys())

    def gorsel(self, girinti:int=None, kademe:int=Decimal('Infinity')) -> None:
        """
        Oluşan Nesneyi Görsel Biçimde Konsola Yazdırır.

        Parametreler
        ------------
            >>> girinti (int) # Kaç Tab Boşluk Bıraksın
            >>> kademe  (int) # En Fazla Kaç Kademe Göstersin

        Kullanım
        --------
        ```
        .gorsel(girinti=0, kademe=1)
        ```

        Dönüt
        -----
        >>> None
        """
        try:
            veri = self.__dict__.copy()
            del veri["_Nesne__sozluk"]

            print(sikici_yaz(veri, girinti, kademe))
        except KeyError:
            print(sikici_yaz(self.__sozluk, girinti, kademe))


class HexCodec:
    @staticmethod
    def encode(utf8_string: str) -> str:
        """UTF-8 stringini kaçış dizileriyle birlikte hex stringine dönüştürür."""
        byte_data   = utf8_string.encode("utf-8")
        hex_string  = byte_data.hex()
        escaped_hex = "\\x".join(hex_string[i:i+2] for i in range(0, len(hex_string), 2))

        return f"\\x{escaped_hex}"

    @staticmethod
    def decode(escaped_hex: str) -> str:
        """Kaçış dizileri içeren bir hex stringini UTF-8 formatındaki stringe dönüştürür."""
        escaped_hex = escaped_hex.strip().replace("\\X", "\\x")

        if isinstance(escaped_hex, str) and not escaped_hex.startswith(r"\x"):
            return escaped_hex.encode("unicode_escape").decode("utf-8")

        hex_string = escaped_hex.replace("\\x", "")
        byte_data  = bytes.fromhex(hex_string)

        return byte_data.decode("utf-8")


# ! Bu araç @keyiflerolsun tarafından | @KekikAkademi için yazılmıştır.

import re

class Packer:
    """
    P.A.C.K.E.R. sıkıştırma ve çözme işlemleri için bir sınıf.
    ! » https://github.com/beautifier/js-beautify/blob/main/python/jsbeautifier/unpackers/packer.py
    """
    @staticmethod
    def clean_escape_sequences(source: str) -> str:
        """Kaçış dizilerini temizler."""
        source = re.sub(r'\\\\', r'\\', source)
        source = source.replace("\\'", "'")
        source = source.replace('\\"', '"')
        return source

    @staticmethod
    def extract_arguments(source: str) -> tuple[str, list[str], int, int]:
        """P.A.C.K.E.R. formatındaki kaynak koddan argümanları çıkarır."""
        match = re.search(r"}\('(.*)',(\d+),(\d+),'(.*)'\.split\('\|'\)", source, re.DOTALL)

        if not match:
            raise ValueError("Invalid P.A.C.K.E.R. source format.")

        payload, radix, count, symtab = match.groups()

        return payload, symtab.split("|"), int(radix), int(count)

    @staticmethod
    def convert_base(s: str, base: int) -> int:
        """Bir sayıyı belirli bir tabandan ondalık tabana çevirir."""
        alphabet = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"

        return sum(alphabet.index(char) * (base**idx) for idx, char in enumerate(reversed(s)))

    @staticmethod
    def lookup_symbol(match: re.Match, symtab: list[str], radix: int) -> str:
        """Sembolleri arar ve yerine koyar."""
        word  = match[0]

        return symtab[Packer.convert_base(word, radix)] or word

    @staticmethod
    def unpack(source: str) -> str:
        """P.A.C.K.E.R. formatındaki sıkıştırılmış bir kaynağı çözer."""
        source = Packer.clean_escape_sequences(source)

        payload, symtab, radix, count = Packer.extract_arguments(source)

        if count != len(symtab):
            raise ValueError("Malformed P.A.C.K.E.R. symtab.")

        return re.sub(r"\b\w+\b", lambda match: Packer.lookup_symbol(match, symtab, radix), payload)

    @staticmethod
    def pack(source: str, radix: int = 62) -> str:
        """Bir metni P.A.C.K.E.R. formatında sıkıştırır."""
        # Bu işlev, simgeleri ve sıkıştırılmış metni yeniden oluşturmak için bir yol sağlar.
        # Ancak bu, belirli bir algoritma veya sıkıştırma tekniğine bağlıdır.
        # Gerçekleştirilmesi zor olabilir çünkü P.A.C.K.E.R.'ın spesifik sıkıştırma mantığını takip etmek gerekir.
        raise NotImplementedError("Packing function is not implemented.")


# ! Bu araç @keyiflerolsun tarafından | @KekikAkademi için yazılmıştır.

from Crypto.Cipher import AES
from Crypto.Random import get_random_bytes
import hashlib, base64

class CryptoJS:
    """
    Paroladan türetilmiş anahtar ile AES/CBC/PKCS7Padding şifreleme ve şifre çözme işlemleri için bir sınıf.
    CryptoJS AES yöntemine uygundur.
    ! » https://gist.github.com/thackerronak/554c985c3001b16810af5fc0eb5c358f
    """
    KEY_SIZE    = 32
    IV_SIZE     = 16
    HASH_CIPHER = "AES/CBC/PKCS7Padding"
    AES_MODE    = AES.MODE_CBC
    KDF_DIGEST  = "md5"
    APPEND      = b"Salted__"

    @staticmethod
    def evp_kdf(password, salt, key_size=32, iv_size=16, iterations=1, hash_algorithm="md5"):
        """Paroladan Anahtar ve IV oluşturmak için bir KDF fonksiyonu."""
        target_key_size = key_size + iv_size
        derived_bytes   = b""
        block           = None

        while len(derived_bytes) < target_key_size:
            hasher = hashlib.new(hash_algorithm)
            if block:
                hasher.update(block)
    
            hasher.update(password)
            hasher.update(salt)
            block = hasher.digest()

            for _ in range(1, iterations):
                block = hashlib.new(hash_algorithm, block).digest()
    
            derived_bytes += block

        return derived_bytes[:key_size], derived_bytes[key_size:key_size + iv_size]

    @staticmethod
    def encrypt(password, plain_text):
        """Verilen metni AES/CBC/PKCS7Padding şifreleme yöntemi ile şifreler."""
        salt    = CryptoJS.generate_salt(8)
        key, iv = CryptoJS.evp_kdf(password.encode("utf-8"), salt, key_size=CryptoJS.KEY_SIZE, iv_size=CryptoJS.IV_SIZE)

        cipher      = AES.new(key, CryptoJS.AES_MODE, iv)
        cipher_text = cipher.encrypt(CryptoJS._pad(plain_text.encode("utf-8")))

        # Create CryptoJS-like encrypted output
        encrypted_data = CryptoJS.APPEND + salt + cipher_text

        return base64.b64encode(encrypted_data).decode("utf-8")

    @staticmethod
    def decrypt(password, cipher_text):
        """Verilen şifreli metni AES/CBC/PKCS7Padding şifreleme yöntemi ile çözer."""
        ct_bytes          = base64.b64decode(cipher_text)
        salt              = ct_bytes[8:16]
        cipher_text_bytes = ct_bytes[16:]

        key, iv = CryptoJS.evp_kdf(password.encode("utf-8"), salt, key_size=CryptoJS.KEY_SIZE, iv_size=CryptoJS.IV_SIZE)

        cipher     = AES.new(key, CryptoJS.AES_MODE, iv)
        plain_text = cipher.decrypt(cipher_text_bytes)

        return CryptoJS._unpad(plain_text).decode("utf-8")

    @staticmethod
    def _pad(s):
        """Veriyi AES blok boyutuna göre doldurur (PKCS7)."""
        block_size = AES.block_size
        padding    = block_size - len(s) % block_size

        return s + bytes([padding] * padding)

    @staticmethod
    def _unpad(s):
        """Dolguyu kaldırır (PKCS7)."""
        return s[:-ord(s[-1:])]

    @staticmethod
    def generate_salt(length):
        """Belirtilen uzunlukta rastgele bir tuz oluşturur."""
        return get_random_bytes(length)


# Bu araç @keyiflerolsun tarafından | @KekikAkademi için yazılmıştır.

from .StringCodec import StringCodec
from .AESManager  import AESManager
from .Packer      import Packer
from .CryptoJS    import CryptoJS
from .HexCodec    import HexCodec
from .NaysHash    import NaysHash


# Bu araç @keyiflerolsun tarafından | @KekikAkademi için yazılmıştır.

from csv    import reader as csv_oku
from typing import List, Dict

def csv2dict(dosya_adi:str) -> List[Dict[str, str]]:
    with open(dosya_adi, "r", encoding="utf-8") as dosya:
        csv_veri = csv_oku(dosya)

        basliklar = next(csv_veri)
        icerik    = list(csv_veri)

    return [dict(zip(basliklar, satir)) for satir in icerik]


# Bu araç @keyiflerolsun tarafından | @KekikAkademi için yazılmıştır.

from .cli         import konsol
from functools    import wraps
from hashlib      import md5
from urllib.parse import urlencode
import time
import threading
import asyncio
import pickle

# Redis client (Redis yoksa fallback yapılacak)
import redis.asyncio as redisAsync
import redis         as redis

# -----------------------------------------------------
# Sabitler ve Yardımcı Fonksiyonlar
# -----------------------------------------------------
UNLIMITED  = None

REDIS_HOST = "127.0.0.1"
REDIS_PORT = 6379
REDIS_DB   = 0
REDIS_PASS = None

def normalize_for_key(value):
    """
    Cache key oluşturma amacıyla verilen değeri normalize eder.
    - Basit tipler (int, float, str, bool, None) doğrudan kullanılır.
    - dict: Anahtarları sıralı olarak normalize eder.
    - list/tuple: Elemanları normalize eder.
    - Diğer: Sadece sınıf ismi kullanılır.
    """
    if isinstance(value, (int, float, str, bool, type(None))):
        return value

    elif isinstance(value, dict):
        return {k: normalize_for_key(value[k]) for k in sorted(value)}

    elif isinstance(value, (list, tuple)):
        return [normalize_for_key(item) for item in value]

    else:
        return value.__class__.__name__

def simple_cache_key(func, args, kwargs) -> str:
    """
    Fonksiyonun tam adı ve parametrelerini kullanarak bir cache key oluşturur.
    Oluşturulan stringin sonuna MD5 hash eklenebilir.
    """
    base_key = f"{func.__module__}.{func.__qualname__}"

    if args:
        norm_args = [normalize_for_key(arg) for arg in args]
        base_key += f"|{norm_args}"

    if kwargs:
        norm_kwargs = {k: normalize_for_key(v) for k, v in kwargs.items()}
        base_key   += f"|{str(sorted(norm_kwargs.items()))}"

    hashed = md5(base_key.encode('utf-8')).hexdigest()
    return f"{base_key}"  # |{hashed}

async def make_cache_key(func, args, kwargs, is_fastapi=False) -> str:
    """
    Cache key'ini oluşturur.
    - is_fastapi=False ise simple_cache_key() kullanılır.
    - True ise FastAPI Request nesnesine göre özel key oluşturulur.
    """
    if not is_fastapi:
        return simple_cache_key(func, args, kwargs)

    # FastAPI: request ilk argümandan ya da kwargs'dan alınır.
    request = args[0] if args else kwargs.get("request")

    if request.method == "GET":
        # Eğer query_params boşsa {} olarak ayarla
        veri = dict(request.query_params) if request.query_params else {}
    else:
        try:
            veri = await request.json()
        except Exception:
            form_data = await request.form()
            veri = dict(form_data.items())

    # Eğer "kurek" gibi özel parametreler varsa temizleyebilirsiniz:
    veri.pop("kurek", None)

    args_hash = md5(urlencode(veri).encode()).hexdigest() if veri else ""
    return f"{request.url.path}?{veri}"

# -----------------------------------------------------
# Senkron Cache (RAM) Sınıfı
# -----------------------------------------------------

class SyncCache:
    """
    Senkron fonksiyonlar için basit in-memory cache.
    TTL süresi dolan veriler periyodik olarak arka plan thread’iyle temizlenir.
    """
    def __init__(self, ttl=UNLIMITED, cleanup_interval=60 * 60):
        self._ttl   = ttl
        self._data  = {}
        self._times = {}

        # TTL sınırsız değilse, cleanup_interval ile ttl'den büyük olanı kullanıyoruz.
        self._cleanup_interval = max(ttl, cleanup_interval) if ttl is not UNLIMITED else cleanup_interval

        # Arka planda çalışan ve periyodik olarak expired entry'leri temizleyen thread başlatılıyor.
        self._lock           = threading.RLock()
        self._cleanup_thread = threading.Thread(target=self._auto_cleanup, daemon=True)
        self._cleanup_thread.start()

    def _auto_cleanup(self):
        """Belirlenen aralıklarla cache içerisindeki süresi dolmuş entry'leri temizler."""
        while True:
            time.sleep(self._cleanup_interval)
            with self._lock:
                keys = list(self._data.keys())
                for key in keys:
                    self.remove_if_expired(key)

    def _is_expired(self, key):
        """Belirtilen key'in süresi dolmuşsa True döner."""
        if self._ttl is UNLIMITED:
            return False

        timestamp = self._times.get(key)

        return timestamp is not None and (time.time() - timestamp > self._ttl)

    def remove_if_expired(self, key):
        """
        Eğer key'in cache süresi dolmuşsa, ilgili entry'yi temizler.
        Thread güvenliği sağlamak için lock kullanılır.
        """
        with self._lock:
            if self._is_expired(key):
                self._data.pop(key, None)
                self._times.pop(key, None)
                # konsol.log(f"[red][-] {key}")

    def __getitem__(self, key):
        with self._lock:
            self.remove_if_expired(key)
            veri = self._data[key]
            # konsol.log(f"[yellow][~] {key}")
            return veri

    def __setitem__(self, key, value):
        with self._lock:
            self._data[key] = value
            if self._ttl is not UNLIMITED:
                self._times[key] = time.time()

class HybridSyncCache:
    """
    Senkron işlemler için, öncelikle Redis cache kullanılmaya çalışılır.
    Redis'ten veri alınamazsa ya da hata oluşursa, SyncCache (in-memory) fallback uygulanır.
    """
    def __init__(self, ttl=None):
        self._ttl = ttl

        try:
            self.redis = redis.Redis(
                host             = REDIS_HOST,
                port             = REDIS_PORT,
                db               = REDIS_DB,
                password         = REDIS_PASS,
                decode_responses = False
            )
            self.redis.ping()
        except Exception:
            self.redis = None

        self.memory = SyncCache(ttl)

    def get(self, key):
        # Önce Redis ile deniyoruz:
        if self.redis:
            try:
                data = self.redis.get(key)
            except Exception:
                data = None
            if data is not None:
                try:
                    result = pickle.loads(data)
                    # konsol.log(f"[yellow][~] {key}")
                    return result
                except Exception:
                    # Deserialize hatası durumunda fallback'e geç
                    pass

        # Redis'te veri yoksa, yerel cache'ten alıyoruz.
        try:
            return self.memory[key]
        except KeyError:
            raise KeyError(key)

    def set(self, key, value):
        try:
            ser = pickle.dumps(value)
        except Exception:
            # Serialization hatası durumunda yerel cache'e yazalım.
            self.memory[key] = value
            return

        if self.redis:
            try:
                if self._ttl is not None:
                    self.redis.set(key, ser, ex=self._ttl)
                else:
                    self.redis.set(key, ser)
                return
            except Exception:
                # Redis'e yazılamazsa yerel cache'e geçelim.
                self.memory[key] = value
                return
        else:
            # Redis kullanılmıyorsa direkt yerel cache'e yaz.
            self.memory[key] = value

    # HybridSyncCache'in subscriptable olması için:
    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        self.set(key, value)

# -----------------------------------------------------
# Asenkron Cache (In-Memory) ve Redis Hybrid Cache
# -----------------------------------------------------

class AsyncCache:
    """
    Temel in-memory asenkron cache.
    """
    def __init__(self, ttl=UNLIMITED, cleanup_interval=60 * 60):
        """
        :param ttl: Her entry için geçerli süre (saniye). Örneğin 3600 saniye 1 saattir.
        :param cleanup_interval: Otomatik temizleme görevinin kaç saniyede bir çalışacağını belirler.
        """
        self._ttl    = ttl
        self._data   = {}
        self._times  = {}
        self.futures = {}

        # TTL sınırsız değilse, cleanup_interval ile ttl'den büyük olanı kullanıyoruz.
        self._cleanup_interval = max(ttl, cleanup_interval) if ttl is not UNLIMITED else cleanup_interval

        # Aktif bir event loop varsa otomatik temizlik görevini başlatıyoruz.
        try:
            self._cleanup_task = asyncio.get_running_loop().create_task(self._auto_cleanup())
        except RuntimeError:
            self._cleanup_task = None

    async def _auto_cleanup(self):
        """Belirlenen aralıklarla cache içerisindeki süresi dolmuş entry'leri temizler."""
        while True:
            await asyncio.sleep(self._cleanup_interval)
            for key in list(self._data.keys()):
                self.remove_if_expired(key)

    def ensure_cleanup_task(self):
        """Event loop mevcutsa, cleanup task henüz başlatılmadıysa oluştur."""
        if self._cleanup_task is None:
            try:
                self._cleanup_task = asyncio.get_running_loop().create_task(self._auto_cleanup())
            except RuntimeError:
                pass  # Yine loop yoksa yapacak bir şey yok

    def remove_if_expired(self, key):
        """
        Belirtilen key'in süresi dolduysa, cache ve futures içerisinden temizler.
        """
        if self._ttl is not UNLIMITED:
            t = self._times.get(key)
            if t is not None and (time.time() - t > self._ttl):
                # konsol.log(f"[red][-] {key}")
                self._data.pop(key, None)
                self._times.pop(key, None)
                self.futures.pop(key, None)

    async def get(self, key):
        """
        Belirtilen key için cache'de saklanan değeri asenkron olarak döndürür.
        Eğer key bulunamazsa, ilgili future üzerinden beklemeyi sağlar.
        """
        # Cleanup task'in aktif olduğundan emin olun.
        self.ensure_cleanup_task()
        # Eğer key'in süresi dolmuşsa, kaldırın.
        self.remove_if_expired(key)

        try:
            # Cache içerisinde key varsa direkt değeri döndür.
            value = self._data[key]
            # konsol.log(f"[yellow][~] {key}")
            return value
        except KeyError:
            # Eğer key cache'de yoksa, aynı key ile başlatılmış future varsa onu bekle.
            future = self.futures.get(key)
            if future:
                await future
                # Future tamamlandığında sonucu döndür.
                value = future.result()
                # konsol.log(f"[yellow][?] {key}")
                return value
            # Eğer future da yoksa, key bulunamadığına dair hata fırlat.
            raise KeyError(key)

    async def set(self, key, value):
        """Belirtilen key için cache'e değer ekler."""
        self.ensure_cleanup_task()
        self._data[key] = value
        if self._ttl is not UNLIMITED:
            self._times[key] = time.time()

class HybridAsyncCache:
    """
    Öncelikle Redis cache kullanılmaya çalışılır.
    Hata durumunda veya Redis erişilemiyorsa in-memory AsyncCache’e geçilir.
    """
    def __init__(self, ttl=UNLIMITED):
        self._ttl = ttl

        try:
            self.redis = redisAsync.Redis(
                host             = REDIS_HOST,
                port             = REDIS_PORT,
                db               = REDIS_DB,
                password         = REDIS_PASS,
                decode_responses = False
            )
        except Exception:
            self.redis = None

        self.memory  = AsyncCache(ttl)
        self.futures = {}

    async def get(self, key):
        # Eşzamanlı istek yönetimi: aynı key için bir future varsa, direkt bekleyip sonucu döndür.
        if key in self.futures:
            # konsol.log(f"[yellow][?] {key}")
            return await self.futures[key]

        # İlk önce Redis ile deneyelim
        if self.redis:
            try:
                data = await self.redis.get(key)
            except Exception:
                return await self.memory.get(key)
            if data is not None:
                try:
                    result = pickle.loads(data)
                    # konsol.log(f"[yellow][~] {key}")
                    return result
                except Exception:
                    # Deserialize hatası durumunda in-memory cache'ten dene
                    return await self.memory.get(key)
            else:
                # Redis'te veri yoksa, in-memory cache'e bak
                return await self.memory.get(key)
        else:
            # Redis kullanılmıyorsa doğrudan in-memory cache'e bak
            return await self.memory.get(key)

    async def set(self, key, value):
        # Önce veriyi pickle etmeyi deniyoruz.
        try:
            ser = pickle.dumps(value)
        except Exception:
            # Serialization hatası durumunda sadece in-memory cache'e yaz
            await self.memory.set(key, value)
            return

        if self.redis:
            try:
                if self._ttl is not UNLIMITED:
                    await self.redis.set(key, ser, ex=self._ttl)
                else:
                    await self.redis.set(key, ser)
                return
            except Exception:
                # Redis yazma hatası durumunda in-memory fallback
                await self.memory.set(key, value)
                return
        else:
            # Redis yoksa in-memory cache'e yaz
            await self.memory.set(key, value)

# -----------------------------------------------------
# Cache'e Hesaplanmış Sonucu Yazma Yardımcı Fonksiyonları
# -----------------------------------------------------

def _sync_maybe_cache(func, key, result, unless):
    """Senkron fonksiyon sonucu için cache kaydı oluşturur."""
    if unless is None or not unless(result):
        func.__cache[key] = result
        # konsol.log(f"[green][+] {key}")

async def _async_compute_and_cache(func, key, unless, *args, **kwargs):
    """
    Asenkron fonksiyon sonucunu hesaplar ve cache'e yazar.
    Aynı key için gelen eşzamanlı çağrılar future üzerinden bekletilir.
    """
    # __cache'den cache nesnesini alıyoruz.
    cache = func.__cache

    # Aynı key için aktif bir future varsa, onun sonucunu döndür.
    if key in cache.futures:
        return await cache.futures[key]

    # Yeni future oluşturuluyor ve cache.futures'e ekleniyor.
    future             = asyncio.Future()
    cache.futures[key] = future

    try:
        # Asenkron fonksiyonu çalıştır ve sonucu elde et.
        result = await func(*args, **kwargs)
        future.set_result(result)

        # unless koşuluna göre cache'e ekleme yap.
        if unless is None or not unless(result):
            await cache.set(key, result)
            # konsol.log(f"[green][+] {key}")

        return result
    except Exception as exc:
        future.cancel()
        raise exc
    finally:
        # İşlem tamamlandığında future'ı temizle.
        cache.futures.pop(key, None)

# -----------------------------------------------------
# kekik_cache Dekoratörü (Senkrondan Asenkrona)
# -----------------------------------------------------

def kekik_cache(ttl=UNLIMITED, unless=None, is_fastapi=False, use_redis=True):
    """
    Bir fonksiyonun (senkron/asenkron) sonucunu cache'ler.
    
    Parametreler:
      - ttl: Cache'in geçerlilik süresi (saniye). UNLIMITED ise süresizdir.
      - unless: Sonuç alınmadan önce çağrılan, True dönerse cache'e alınmaz.
      - is_fastapi: True ise, FastAPI Request nesnesine göre key oluşturur.
      - use_redis: Asenkron fonksiyonlarda Redis kullanımı (Hybrid cache) için True verilebilir.
    
    Örnek Kullanım:
    
        @kekik_cache(ttl=15, unless=lambda sonuc: sonuc is None)
        async def bakalim(param):
            return param
    """
    # Parametresiz kullanım durumunda
    if callable(ttl):
        return kekik_cache(UNLIMITED, unless=unless, is_fastapi=is_fastapi, use_redis=use_redis)(ttl)

    def decorator(func):
        if asyncio.iscoroutinefunction(func):
            # Asenkron fonksiyonlar için cache türünü seçelim:

            func.__cache = HybridAsyncCache(ttl) if use_redis else AsyncCache(ttl)

            @wraps(func)
            async def async_wrapper(*args, **kwargs):
                key = await make_cache_key(func, args, kwargs, is_fastapi)

                try:
                    return await func.__cache.get(key)
                except KeyError:
                    return await _async_compute_and_cache(func, key, unless, *args, **kwargs)

            return async_wrapper
        else:
            # Senkron fonksiyonlar için
            func.__cache = HybridSyncCache(ttl) if use_redis else SyncCache(ttl)

            @wraps(func)
            def sync_wrapper(*args, **kwargs):
                key = simple_cache_key(func, args, kwargs)

                try:
                    return func.__cache[key]
                except KeyError:
                    result = func(*args, **kwargs)
                    _sync_maybe_cache(func, key, result, unless)
                    return result

            return sync_wrapper

    return decorator